# Generated by Django 3.1.4 on 2020-12-16 07:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0025_auto_20201216_0917"),
    ]

    operations = [
        migrations.AddField(
            model_name="courserun",
            name="sync_mode",
            field=models.CharField(
                choices=[
                    ("manual", "Manual"),
                    ("sync_to_draft", "Synchronization to draft page"),
                    ("sync_to_public", "Synchronization to public page"),
                ],
                default="manual",
                max_length=20,
            ),
        ),
    ]



"""Custom template tags for the courses application of Richie."""
import json

from django import template
from django.core.exceptions import ObjectDoesNotExist
from django.template.loader import render_to_string
from django.utils import timezone
from django.utils.translation import get_language
from django.utils.translation import gettext as _
from django.utils.translation import to_locale

import arrow
from classytags.arguments import Argument, MultiValueArgument
from classytags.core import Options, Tag
from classytags.utils import flatten_context
from cms.templatetags.cms_tags import (
    Placeholder,
    PlaceholderOptions,
    _get_page_by_untyped_arg,
)
from cms.toolbar.utils import get_toolbar_from_request
from cms.utils import get_site_id
from cms.utils.plugins import get_plugins

from ..lms import LMSHandler

# pylint: disable=invalid-name
register = template.Library()


def get_plugins_render_tag(context, name, varname, nodelist, page_lookup=None):
    """
    Retrieve the placeholder's plugins and set them as a variable in the template context.
    If the placeholder is empty, render the block as fallback content and return the
    resulting HTML.
    If the placeholder is editable and rendered on its own page, the edit script and markup
    are added to the HTML content.
    """
    content = ""
    request = context.get("request")

    if request:

        context[varname] = []
        page = _get_page_by_untyped_arg(page_lookup, request, get_site_id(None))

        if not page:
            return ""

        try:
            placeholder = page.placeholders.get(slot=name)
        except ObjectDoesNotExist:
            return ""
        else:
            context[varname] = [
                cms_plugin.get_plugin_instance()[0]
                for cms_plugin in get_plugins(
                    request, placeholder, template=page.get_template()
                )
            ]

        # Default content if there is no plugins in the placeholder
        if not context[varname] and nodelist:
            content = nodelist.render(context)

        # Add the edit script and markup to the content, only if the placeholder is editable
        # and the visited page is the one on which the placeholder is declared.
        toolbar = get_toolbar_from_request(request)
        if placeholder.page == request.current_page and toolbar.edit_mode_active:
            renderer = toolbar.get_content_renderer()
            data = renderer.get_editable_placeholder_context(placeholder, page=page)
            data["content"] = content
            content = renderer.placeholder_edit_template.format(**data)

    return content


@register.tag("get_placeholder_plugins")
class GetPlaceholderPlugins(Placeholder):
    """
    A template tag that declares a placeholder and sets its plugins as a context variable
    instead of rendering them eg:

        {% get_placeholder_plugins "logo" as varname %}
        {% get_placeholder_plugins "logo" as varname or %}
            <div>No content</div>
        {% endget_placeholder_plugins %}

    This tag can typically be used in association with the block_plugin tag, to customize the
    way it is rendered eg:

        {% get_placeholder_plugins "logo" as plugins %}
        {% blockplugin plugins.0 %}
            <img src="{% thumbnail instance.picture 300x150 %}"/>
        {% endblockplugin %}

    Keyword arguments:
        name: the name of the placeholder
        varname: context variable name. Output will be added to template context as this variable
            instead of being returned.
        or: optional argument which if given will make the template tag a block
            tag whose content is shown if the placeholder is empty

    Note: We must derive from the Placeholder class so that the tag is recognized as a
          placeholder and shown in the structure toolbar.
    """

    name = "get_placeholder_plugins"
    options = PlaceholderOptions(
        Argument("name", resolve=False),
        "as",
        Argument("varname", resolve=False),
        MultiValueArgument("extra_bits", required=False, resolve=False),
        blocks=[("endget_placeholder_plugins", "nodelist")],
    )

    # pylint: disable=arguments-differ,too-many-arguments
    def render_tag(self, context, name, varname, extra_bits, nodelist=None):
        return get_plugins_render_tag(context, name, varname, nodelist)


@register.tag("get_page_plugins")
class GetPagePlugins(Tag):
    """
    A template tag that gets plugins from a page's placeholder returns them as a context variable:

        {% get_page_plugins "logo" page_lookup as varname %}
        {% get_page_plugins "logo" page_lookup as varname or %}
            <div>No content</div>
        {% endget_page_plugins %}

    This tag can typically be used in association with the block_plugin tag,
    to render the retrieved plugins:

        {% get_page_plugins "logo" page_lookup as plugins %}
        {% blockplugin plugins.0 %}
            <img src="{% thumbnail instance.picture 300x150 %}"/>
        {% endblockplugin %}

    Keyword arguments:
        name: the name of the placeholder
        page_lookup: lookup argument for Page. See `_get_page_by_untyped_arg()`
            for detailed information on the allowed types and their interpretation for the
            `page_lookup` argument.
        varname: context variable name. Output will be added to template context as this variable
            instead of being returned.
        or: optional argument which if given will make the template tag a block
            tag whose content is shown if the placeholder is empty
    """

    name = "get_page_plugins"
    options = PlaceholderOptions(
        Argument("name", resolve=False),
        Argument("page_lookup"),
        "as",
        Argument("varname", resolve=False),
        MultiValueArgument("extra_bits", required=False, resolve=False),
        blocks=[("endget_page_plugins", "nodelist")],
    )

    # pylint: disable=arguments-differ,too-many-arguments, unused-argument
    def render_tag(
        self, context, name, page_lookup, varname, extra_bits, nodelist=None
    ):
        return get_plugins_render_tag(context, name, varname, nodelist, page_lookup)


@register.tag()
class BlockPlugin(Tag):
    """
    Like DjangoCMS 'render_plugin_block' but only includes the edit script and markup when
    the related placeholder is editable.

    This issue was raised to DjangoCMS and we need our own template tag until they find a way
    to fix it in DjangoCMS (https://github.com/divio/django-cms/issues/6683).
    """

    name = "blockplugin"
    template = "cms/toolbar/plugin.html"
    options = Options(Argument("plugin"), blocks=[("endblockplugin", "nodelist")])

    # pylint: disable=arguments-differ
    def render_tag(self, context, plugin, nodelist):
        """
        Renders the block for the plugin and returns the resulting HTML leaving the temmpate
        context untouched.
        If the placholder is editable, the edit script and markup are added to the rendered HTML.
        """
        request = context.get("request")
        if not plugin or not request:
            return ""

        # Add the plugin and its rendered content to an internal context
        internal_context = flatten_context(context)
        internal_context["instance"] = plugin
        internal_context["content"] = nodelist.render(context.new(internal_context))

        # Add the edit script and markup to the content, only if the placeholder is editable
        # and the visited page is the one on which the plugin's placeholder is declared.
        toolbar = get_toolbar_from_request(request)
        if plugin.placeholder.page == request.current_page and toolbar.edit_mode_active:
            return render_to_string(self.template, internal_context)

        return internal_context["content"]


@register.filter()
def is_empty_placeholder(page, slot):
    """A template filter to determine if a placeholder is empty.

    This is useful when we don't want to include any wrapper markup in our template unless
    the placeholder unless it actually contains plugins.
    """
    placeholder = page.placeholders.get(slot=slot)
    return not placeholder.cmsplugin_set.exists()


@register.filter()
def order_by(queryset, args):
    """A template filter to force ordering on a queryset.

    Taken from: https://djangosnippets.org/snippets/741/
    This is useful for DjangoCMS page querysets because we don't have access to the view.
    """
    args = [x.strip() for x in args.split(",")]
    return queryset.order_by(*args)


@register.filter()
def has_connected_lms(course_run):
    """
    Determine if the passed course run has a connected LMS (as determined through out LMSHandler
    and settings).
    This enables our templates to either use the <CourseRunEnrollment /> component or a simple
    link to the course run.
    """
    return LMSHandler.select_lms(course_run.resource_link) is not None


@register.simple_tag(takes_context=True)
def course_enrollment_widget_props(context):
    """
    Return a json dumps which contains all course_run's properties required by
    CourseEnrollment React widget
    """
    course_run = context["run"]

    profile_urls = json.loads(
        context.get("AUTHENTICATION", {}).get("profile_urls", "{}")
    )
    dashboard_link = profile_urls.get("dashboard", {}).get("action")

    starts_in_message = None
    if course_run.start > timezone.now():
        course_start = arrow.get(course_run.start)
        humanized_course_start = course_start.humanize(
            arrow.now(), locale=to_locale(get_language())
        )
        # Translators: delay indicates when the course will start as a duration.
        # In english the string will be "The course will start in 3 days"
        starts_in_message = _("The course will start {delay:s}").format(
            delay=humanized_course_start
        )

    return json.dumps(
        {
            "courseRun": {
                "id": course_run.id,
                "resource_link": course_run.resource_link,
                "priority": course_run.state["priority"],
                "starts_in_message": starts_in_message,
                "dashboard_link": dashboard_link,
            }
        }
    )



# Generated by Django 2.1.5 on 2019-02-09 08:29

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import filer.fields.image


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("cms", "0022_auto_20180620_1551"),
        migrations.swappable_dependency(settings.FILER_IMAGE_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="LargeBanner",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="large_banner_largebanner",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                ("title", models.CharField(max_length=255)),
                ("logo_alt_text", models.CharField(max_length=255)),
                (
                    "background_image",
                    filer.fields.image.FilerImageField(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="background_image",
                        to=settings.FILER_IMAGE_MODEL,
                        verbose_name="background image",
                    ),
                ),
                (
                    "logo",
                    filer.fields.image.FilerImageField(
                        on_delete=django.db.models.deletion.PROTECT,
                        related_name="logo",
                        to=settings.FILER_IMAGE_MODEL,
                        verbose_name="logo",
                    ),
                ),
            ],
            options={"abstract": False},
            bases=("cms.cmsplugin",),
        )
    ]



"""
NestedItem CMS plugin
"""
from django.utils.translation import gettext_lazy as _

from cms.plugin_base import CMSPluginBase
from cms.plugin_pool import plugin_pool

from richie.apps.core.defaults import PLUGINS_GROUP

from .defaults import NESTEDITEM_VARIANTS
from .models import NestedItem


@plugin_pool.register_plugin
class NestedItemPlugin(CMSPluginBase):
    """
    CMSPlugin to add a nested item with a variant form factor
    """

    cache = True
    model = NestedItem
    module = PLUGINS_GROUP
    name = _("Nested item")
    allow_children = True
    render_template = "richie/nesteditem/nesteditem.html"
    fieldsets = ((None, {"fields": ["variant", "content"]}),)

    def render(self, context, instance, placeholder):
        context.update(
            {
                "instance": instance,
                "placeholder": placeholder,
                "nesting_level": context.get("nesting_level", 0),
                "parent_variant": context.get(
                    "parent_variant", NESTEDITEM_VARIANTS[0][0]
                ),
            }
        )
        return context



# Generated by Django 2.2.8 on 2020-01-02 13:53

from django.db import migrations, models

import richie.apps.core.fields.multiselect


class Migration(migrations.Migration):

    dependencies = [("courses", "0010_auto_footer_to_static_placeholder")]

    operations = [
        migrations.RenameField(
            model_name="licence", old_name="content", new_name="content_deprecated"
        ),
        migrations.RenameField(
            model_name="licence", old_name="name", new_name="name_deprecated"
        ),
        migrations.AlterField(
            model_name="licence",
            name="content_deprecated",
            field=models.TextField(
                db_column="content", default="", verbose_name="content"
            ),
        ),
        migrations.AlterField(
            model_name="licence",
            name="name_deprecated",
            field=models.CharField(
                db_column="name", max_length=200, verbose_name="name"
            ),
        ),
    ]



"""
Declare and configure the models for the courses application
"""
# pylint: disable=too-many-lines
from collections.abc import Mapping
from datetime import MAXYEAR, datetime

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.db.models import Q
from django.urls import reverse
from django.utils import timezone
from django.utils.functional import cached_property, lazy
from django.utils.translation import gettext_lazy as _

import pytz
from cms.constants import PUBLISHER_STATE_DIRTY
from cms.extensions.extension_pool import extension_pool
from cms.models import Page, PagePermission
from cms.models.pluginmodel import CMSPlugin
from filer.fields.image import FilerImageField
from filer.models import FolderPermission
from parler.models import TranslatableModel, TranslatedField, TranslatedFieldsModel

from ...core.defaults import ALL_LANGUAGES
from ...core.fields.duration import CompositeDurationField
from ...core.fields.multiselect import MultiSelectField
from ...core.helpers import get_permissions
from ...core.models import BasePageExtension
from .. import defaults
from .category import Category, CategoryPluginModel
from .organization import Organization, OrganizationPluginModel
from .person import Person, PersonPluginModel
from .role import PageRole

MAX_DATE = datetime(MAXYEAR, 12, 31, tzinfo=pytz.utc)


class CourseState(Mapping):
    """An immutable object to describe a course (resp. course run) state."""

    (
        ONGOING_OPEN,
        FUTURE_OPEN,
        ARCHIVED_OPEN,
        FUTURE_NOT_YET_OPEN,
        FUTURE_CLOSED,
        ONGOING_CLOSED,
        ARCHIVED_CLOSED,
        TO_BE_SCHEDULED,
    ) = range(8)

    STATE_CALLS_TO_ACTION = {
        ONGOING_OPEN: _("enroll now"),
        FUTURE_OPEN: _("enroll now"),
        ARCHIVED_OPEN: _("study now"),
        FUTURE_NOT_YET_OPEN: None,
        FUTURE_CLOSED: None,
        ONGOING_CLOSED: None,
        ARCHIVED_CLOSED: None,
        TO_BE_SCHEDULED: None,
    }

    STATE_TEXTS = {
        ONGOING_OPEN: _("closing on"),
        FUTURE_OPEN: _("starting on"),
        ARCHIVED_OPEN: _("closing on"),
        FUTURE_NOT_YET_OPEN: _("starting on"),
        FUTURE_CLOSED: _("enrollment closed"),
        ONGOING_CLOSED: _("on-going"),
        ARCHIVED_CLOSED: _("archived"),
        TO_BE_SCHEDULED: _("to be scheduled"),
    }

    def __init__(self, priority, date_time=None):
        """
        Initialize a course state with a priority and optionally a datetime.

        Several states are possible for a course run each of which is given a priority. The
        lower the priority, the more interesting the course run is (a course run open for
        enrollment is more interesting than an archived course run):
        - ONGOING_OPEN: a run is on-going and open for enrollment > "closing on": {enrollment_end}
        - FUTURE_OPEN: a run is future and open for enrollment > "starting on": {start}
        - ARCHIVED_OPEN: a run is past but open for enrollment > "closing on": {enrollment_end}
        - FUTURE_NOT_YET_OPEN: a run is future and not yet open for enrollment
            > "starting on": {start}
        - FUTURE_CLOSED: a run is future and no more open for enrollment > "closed": {None}
        - ONGOING_CLOSED: a run is on-going but closed for enrollment > "on going": {None}
        - ARCHIVED_CLOSED: there's a finished run in the past > "archived": {None}
        - TO_BE_SCHEDULED: theres's no run with a start date or no run at all
            > "to be scheduled": {None}
        """
        # Check that `date_time` is set when it should be
        if date_time is None and priority in [
            CourseState.ONGOING_OPEN,
            CourseState.FUTURE_OPEN,
            CourseState.ARCHIVED_OPEN,
            CourseState.FUTURE_NOT_YET_OPEN,
        ]:
            raise ValidationError(
                "date_time should not be null for a {:d} course state.".format(priority)
            )

        # A special case of being open is when enrollment never ends
        text = self.STATE_TEXTS[priority]
        if (
            priority in [CourseState.ONGOING_OPEN, CourseState.ARCHIVED_OPEN]
            and date_time.year == MAXYEAR
        ):
            text = _("forever open")
            date_time = None
        kwargs = {
            "priority": priority,
            "datetime": date_time,
            "call_to_action": self.STATE_CALLS_TO_ACTION[priority],
            "text": text,
        }
        self._d = dict(**kwargs)

    def __iter__(self):
        """Iterate on the inner dictionary."""
        return iter(self._d)

    def __len__(self):
        """Return length of the inner dictionary."""
        return len(self._d)

    def __getitem__(self, key):
        """Access the inner dictionary."""
        return self._d[key]

    def __lt__(self, other):
        """Make it easy to compare two course states."""
        return self._d["priority"] < other["priority"]


class Course(BasePageExtension):
    """
    The course page extension represents and records a course in the catalog.

    This model should be used to record structured data about the course whereas the
    associated page object is where we record the less structured information to display on the
    page that presents the course.
    """

    CODE_MAX_LENGTH = 100

    code = models.CharField(
        _("code"),
        max_length=CODE_MAX_LENGTH,
        help_text=_("Unique reference for the course."),
        null=True,
        blank=True,
    )

    duration = CompositeDurationField(
        time_units=defaults.TIME_UNITS,
        default_unit=defaults.DEFAULT_TIME_UNIT,
        max_length=80,
        blank=True,
        null=True,
        help_text=_("The course time range."),
    )

    effort = CompositeDurationField(
        time_units=defaults.EFFORT_UNITS,
        default_unit=defaults.DEFAULT_EFFORT_UNIT,
        max_length=80,
        blank=True,
        null=True,
        help_text=_("Total amount of time to complete this course."),
    )

    is_listed = models.BooleanField(
        default=True,
        verbose_name=_("is listed"),
        help_text=_("Tick if you want the course to be visible on the search page."),
    )

    is_self_paced = models.BooleanField(
        default=False,
        verbose_name=_("is self paced"),
        help_text=_("Tick if the course pace is self paced."),
    )

    PAGE = defaults.COURSES_PAGE

    class Meta:
        db_table = "richie_course"
        verbose_name = _("course")
        verbose_name_plural = _("courses")

    def __str__(self):
        """Human representation of a course."""
        return "{model}: {title}".format(
            model=self._meta.verbose_name.title(),
            title=self.extended_object.get_title(),
        )

    def validate_unique(self, exclude=None):
        """
        We can't enforce code unicity with a db constraint because it is repeated across
        the draft and public versions of the page AND in each snapshot of a course.

        A complex unique constraint would have to span the "extended_object" public key
        which is not possible.
        """
        super().validate_unique()
        if (
            self.extended_object.publisher_is_draft
            and not self.is_snapshot
            and self.__class__.objects.exclude(pk=self.pk)
            .filter(
                code=self.code,
                # allow public duplicates
                extended_object__publisher_is_draft=True,
                # allow snapshot duplicates
                extended_object__node__parent__cms_pages__course__isnull=True,
            )
            .exists()
        ):
            raise ValidationError({"code": ["Code must be unique."]})

    def get_admin_url_to_add_run(self, request):
        """
        Get the admin url of the form to add a course run with the "direct_course"
        field pre-populated to self.
        """
        base_url = reverse("admin:courses_courserun_add")
        return f"{base_url:s}?direct_course={self.id:d}"

    @property
    def is_snapshot(self):
        """Return True if the course is a snapshot (it has a course as parent)."""
        # Look for the page and not the course as the course may not yet be created in database
        # (see "validate_unique" method below)
        return Page.objects.filter(
            id=self.extended_object.id, node__parent__cms_pages__course__isnull=False
        ).exists()

    def create_page_role(self):
        """
        Create a new page role for the course with:
          - a user group to handle permissions for admins of this course,
          - a folder in Django Filer to store images related to this course,
          - all necessary permissions.
        """
        if not getattr(settings, "CMS_PERMISSION", False):
            return None

        # The page role is only created for draft courses
        if not self.extended_object.publisher_is_draft:
            return None

        # Don't do anything if it already exists
        page_role = PageRole.objects.filter(
            page=self.extended_object, role=defaults.ADMIN
        ).first()

        if page_role:
            return page_role

        # Create a role for admins of this course (which will automatically create a new
        # user group and a new Filer folder)
        page_role = PageRole.objects.create(
            page=self.extended_object, role=defaults.ADMIN
        )

        # Associate permissions as defined in settings:
        # - Create Django permissions
        page_role.group.permissions.set(
            get_permissions(defaults.COURSE_ADMIN_ROLE.get("django_permissions", []))
        )

        # - Create DjangoCMS page permissions
        PagePermission.objects.create(
            group_id=page_role.group_id,
            page=self.extended_object,
            **defaults.COURSE_ADMIN_ROLE.get("course_page_permissions", {}),
        )

        # - Create the Django Filer folder permissions
        FolderPermission.objects.create(
            folder_id=page_role.folder_id,
            group_id=page_role.group_id,
            **defaults.COURSE_ADMIN_ROLE.get("course_folder_permissions", {}),
        )
        return page_role

    def create_permissions_for_organization(self, organization):
        """
        Create page and folder permissions on the course page for the admin group of the
        organization passed in argument.
        """
        course_page_role = self.create_page_role()
        organization_page_role = organization.create_page_role()

        if organization_page_role is None or course_page_role is None:
            return

        # - Create DjangoCMS page permissions
        PagePermission.objects.get_or_create(
            group_id=organization_page_role.group_id,
            page_id=self.extended_object_id,
            defaults=defaults.ORGANIZATION_ADMIN_ROLE.get(
                "courses_page_permissions", {}
            ),
        )

        # - Create the Django Filer folder permissions
        FolderPermission.objects.get_or_create(
            group_id=organization_page_role.group_id,
            folder_id=course_page_role.folder_id,
            defaults=defaults.ORGANIZATION_ADMIN_ROLE.get(
                "courses_folder_permissions", {}
            ),
        )

    def get_organizations(self, language=None):
        """
        Return the organizations linked to the course via an organization plugin in any
        of the placeholders on the course detail page, ranked by their `path` to respect
        the order in the organizations tree.
        """
        return self.get_direct_related_page_extensions(
            Organization, OrganizationPluginModel, language=language
        )

    def get_main_organization(self):
        """
        Return the main organization linked to the course via an organization plugin on the
        course detail page.

        Plugins are sortable by drag-and-drop in the plugin bar so we use this "position"
        information to return the first one as the main organization. We assume that only one
        placeholder holds organization plugins and return the first organization in order of
        position in this placeholder.
        """
        return (
            self.get_organizations()
            .order_by("extended_object__organization_plugins__cmsplugin_ptr__position")
            .first()
        )

    def get_persons(self, language=None):
        """
        Return the persons linked to the course via a person plugin in any of the
        placeholders on the course detail page, ranked by their `path` to respect
        the order in the persons tree.
        """
        return self.get_direct_related_page_extensions(
            Person, PersonPluginModel, language=language
        )

    def get_categories(self, language=None):
        """
        Return the categories linked to the course via a category plugin in any of the
        placeholders on the course detail page, ranked by their `path` to respect the
        order in the categories tree.
        """
        return self.get_direct_related_page_extensions(
            Category, CategoryPluginModel, language=language
        )

    def get_root_to_leaf_category_pages(self):
        """
        Build a query retrieving all pages linked to this course's categories or one of their
        ancestors excluding the meta category itself.

        This is useful to build the course Elasticsearch index. When a category is linked to a
        course, we associate all its ancestors (excluding the meta category) to the course so
        that a search for courses in an ancestor category also returns the course.
        """
        category_query = self.get_categories()

        # 1. We want the pages directly related to a category of the course
        page_query = Page.objects.filter(publisher_draft__category__in=category_query)
        # 2. We want the pages related to one of the ancestors of the categories of the course
        for category in category_query.select_related(
            "public_extension__extended_object"
        ):
            ancestor_query = (
                category.public_extension.extended_object.get_ancestor_pages()
            )
            # Don't include the meta category as it materializes a "filter bank" and not a
            # search option
            page_query = page_query | ancestor_query.filter(
                node__parent__cms_pages__category__isnull=False
            )
        return page_query.distinct()

    def get_snapshots(self, include_self=False):
        """
        Get a query yielding all snapshots of a course.

        Arguments:
        ----------
        include_self (boolean): whether the current object should be included in the query
            or just its snapshots.

        Returns:
        --------
        Queryset[Course]: All snapshots of a course (calling this method on a snapshot would
            return an empty queryset)

        """
        is_draft = self.extended_object.publisher_is_draft
        node = self.extended_object.node
        current_and_descendant_nodes = node.__class__.get_tree(parent=node)

        query = self.__class__.objects.filter(
            extended_object__node__in=current_and_descendant_nodes,
            extended_object__publisher_is_draft=is_draft,
        )

        if include_self is False:
            query = query.exclude(pk=self.pk)

        return query

    def get_course_runs(self):
        """
        Returns a query yielding the course runs related to the course. They may be directly
        related to the course or to a snapshot of the course.

        The draft and the public page have their own course runs. The course runs on the draft
        page are copied to the public page when the course is published (for any language).
        """
        is_draft = self.extended_object.publisher_is_draft
        node = self.extended_object.node
        current_and_descendant_nodes = node.__class__.get_tree(parent=node)

        return CourseRun.objects.filter(
            direct_course__extended_object__node__in=current_and_descendant_nodes,
            direct_course__extended_object__publisher_is_draft=is_draft,
        ).order_by("-start")

    @cached_property
    def course_runs_dict(self):
        """Returns a dict of course runs grouped by their state."""
        course_runs_dict = {
            i: [] for i in range(len(CourseState.STATE_CALLS_TO_ACTION))
        }
        for run in self.get_course_runs():
            course_runs_dict[run.state["priority"]].append(run)

        return dict(course_runs_dict)

    def get_pace(self):
        """
        Returns the course pace computed according to duration and effort fields or
        None if one of these fields is blank or if `is_self_paced` flag is equal to False.

        This method returns a tuple of three values:
        - pace : the float value of the pace (ratio between effort / duration)
        - pace_unit
        - pace_reference_unit
        """
        if self.is_self_paced is True or self.duration is None:
            return None

        if self.effort is None:
            raise ValueError("Cannot compute pace without effort.")

        (effort, effort_unit) = self.effort
        (duration, duration_unit) = self.duration

        time_units_keys = list(dict(defaults.TIME_UNITS))
        if time_units_keys.index(effort_unit) >= time_units_keys.index(duration_unit):
            raise ValueError(
                (
                    "Cannot compute pace with effort unit less than or equal to the duration unit."
                    f" ({effort_unit}/{duration_unit})"
                )
            )

        pace = effort / duration
        return (pace, effort_unit, duration_unit)

    def get_pace_display(self):
        """
        Returns a human readable pace value to display.
        If the course is self paced or has no duration, it returns `Self paced`.
        If pace cannot be computed, it returns None.

        To improve readability, we manage three cases with different renders:
        - Pace is a full hour -> ~2 hours/day
        - Pace is more than a full hour -> ~2h30/day
        - Pace is under an hour -> ~45 minutes/day

        Furthermore, as this information is an estimation, we do not want the
        value be too precise, so we round minutes to fifteen minutes slots.
        """
        try:
            (pace, pace_unit, pace_reference_unit) = self.get_pace()
        except TypeError:
            return _("Self paced")
        except ValueError:
            return None

        time_units = defaults.TIME_UNITS
        pace_in_minutes = pace * 60 if pace_unit == defaults.HOUR else pace
        pace_hours = int(pace_in_minutes // 60)
        pace_minutes = int(pace_in_minutes % 60)
        pace_minutes_rounded_by_quarter = (
            round(pace_minutes / 15) * 15
        )  # round minutes by fifteen

        # Pace is under an hour
        if pace_hours == 0:
            resolution = 15 if pace_minutes > 15 else 5
            return _("~{pace:d} {effort_unit!s}/{duration_unit!s}").format(
                pace=round(pace_minutes / resolution) * resolution
                or 5,  # Display at least 5 minutes
                effort_unit=time_units[defaults.MINUTE][1],
                duration_unit=time_units[pace_reference_unit][0],
            )
        # Pace is a full hour
        if pace_minutes_rounded_by_quarter == 0:
            count_index = 1 if pace_hours > 1 else 0
            return _("~{pace:d} {effort_unit!s}/{duration_unit!s}").format(
                pace=pace_hours,
                effort_unit=time_units[defaults.HOUR][count_index],
                duration_unit=time_units[pace_reference_unit][0],
            )

        # Pace is more than a full hour
        return _("~{pace_hours:d}h{pace_minutes:d}/{duration_unit!s}").format(
            pace_hours=pace_hours,
            pace_minutes=pace_minutes_rounded_by_quarter,
            duration_unit=time_units[pace_reference_unit][0],
        )

    def get_programs(self, language=None):
        """
        Return a query to get the programs related to this course ie for which a plugin for
        this course is linked to the program page via any placeholder.
        """
        return self.get_reverse_related_page_extensions("program", language=language)

    @property
    def state(self):
        """
        The state of the course carrying information on what to display on a course glimpse.

        The game is to find the highest priority state for this course among its course runs.
        """
        # The default state is for a course that has no course runs
        best_state = CourseState(CourseState.TO_BE_SCHEDULED)

        for course_run in self.get_course_runs().only(
            "start", "end", "enrollment_start", "enrollment_end"
        ):
            state = course_run.state
            if state < best_state:
                best_state = state
            if state["priority"] == CourseState.ONGOING_OPEN:
                # We found the best state, don't waste more time
                break

        return best_state

    def copy_relations(self, oldinstance, language=None):
        """
        This method is called for 2 types of copying:
            1- cloning (copying a draft page to another position in the page tree)
            2- publishing (copying a draft page to its public counterpart)

        When cloning a course, we do not want the related course runs to follow. Each copy of the
        course has its own distinct runs.

        When publishing a course, the course runs must be manually copied to the public page as
        it should perfectly reflect the state of the draft page at the moment we clicked on the
        publish button.
        """
        # Don't copy the course runs if this is a copy to clone
        if oldinstance.public_extension_id != self.id:
            return

        # Since this is a copy to publish, let's copy the course runs
        for old_course_run in oldinstance.runs.all():
            course_run = CourseRun.objects.get(pk=old_course_run.pk)
            course_run.direct_course = self

            # If the course run already exists on the public course, update it,
            try:
                public_course_run = old_course_run.public_course_run
            except CourseRun.DoesNotExist:
                # Otherwise create a new public course run and link it to its draft
                course_run.draft_course_run_id = old_course_run.id
                course_run.pk = None
                course_run.save()
            else:
                course_run.draft_course_run_id = old_course_run.id
                course_run.pk = public_course_run.pk
                course_run.save()

            course_run.copy_translations(old_course_run, language=language)

    def save(self, *args, **kwargs):
        """
        Enforce validation each time an instance is saved
        """
        self.full_clean()
        super().save(*args, **kwargs)


class CourseRunSyncMode(models.TextChoices):
    """Course run synchronization mode choices for the "sync_mode" field."""

    MANUAL = "manual", _("Manual")
    SYNC_TO_DRAFT = "sync_to_draft", _("Synchronization to draft page")
    SYNC_TO_PUBLIC = "sync_to_public", _("Synchronization to public page")


class CourseRun(TranslatableModel):
    """
    The course run represents and records the occurence of a course between a start
    and an end date.
    """

    direct_course = models.ForeignKey(
        Course, on_delete=models.CASCADE, related_name="runs"
    )
    # We register the foreign key in "draft_course_run" and not in "public_course_run"
    # so that the public course run gets deleted by cascade in the database when the
    # draft page is deleted. Doing it the other way would be fragile.
    draft_course_run = models.OneToOneField(
        "self",
        on_delete=models.CASCADE,
        null=True,
        editable=False,
        related_name="public_course_run",
    )
    sync_mode = models.CharField(
        max_length=20,
        choices=CourseRunSyncMode.choices,
        default=CourseRunSyncMode.MANUAL,
    )

    title = TranslatedField()
    resource_link = models.CharField(
        _("resource link"), max_length=200, blank=True, null=True
    )
    start = models.DateTimeField(_("course start"), blank=True, null=True)
    end = models.DateTimeField(_("course end"), blank=True, null=True)
    enrollment_start = models.DateTimeField(
        _("enrollment start"), blank=True, null=True
    )
    enrollment_end = models.DateTimeField(_("enrollment end"), blank=True, null=True)
    languages = MultiSelectField(
        max_choices=50,
        max_length=255,  # MySQL does not allow max_length > 255
        # Language choices are made lazy so that we can override them in our tests.
        # When set directly, they are evaluated too early and can't be changed with the
        # "override_settings" utility.
        choices=lazy(lambda: ALL_LANGUAGES, tuple)(),
        help_text=_("The list of languages in which the course content is available."),
    )

    class Meta:
        db_table = "richie_course_run"
        verbose_name = _("course run")
        verbose_name_plural = _("course runs")

    def __str__(self):
        """Human representation of a course run."""
        start = "{:%y/%m/%d %H:%M} - ".format(self.start) if self.start else ""
        return f"Course run {self.id!s} starting {start:s}"

    def copy_translations(self, oldinstance, language=None):
        """Copy translation objects for a language if provided or for all languages."""
        query = CourseRunTranslation.objects.filter(master=oldinstance)
        if language:
            query = query.filter(language_code=language)

        for translation_object in query:
            try:
                target_pk = CourseRunTranslation.objects.filter(
                    master=self, language_code=translation_object.language_code
                ).values_list("pk", flat=True)[0]
            except IndexError:
                translation_object.pk = None
            else:
                translation_object.pk = target_pk
            translation_object.master = self
            translation_object.save()

    def mark_course_dirty(self):
        """
        Mark the related course page as dirty if the course run has changed since it was last
        published, so that the modifications can be checked and confirmed by a reviewer.
        """
        try:
            public_instance = self.__class__.objects.get(draft_course_run__pk=self.pk)
        except self.__class__.DoesNotExist:
            # This is a new instance, mark page dirty in all languages unless
            # the course run is yet to be scheduled (hidden from public page in this case)
            if self.state["priority"] < CourseState.TO_BE_SCHEDULED:
                self.direct_course.extended_object.title_set.update(
                    publisher_state=PUBLISHER_STATE_DIRTY
                )
            return

        is_visible = (
            self.state["priority"] < CourseState.TO_BE_SCHEDULED
            or public_instance.state["priority"] < CourseState.TO_BE_SCHEDULED
        )
        # Mark the related course page dirty if the course run content has changed
        # Break out of the for loop as soon as we found a difference
        for field in self._meta.fields:
            if field.name == "direct_course":
                if (
                    public_instance.direct_course.draft_extension != self.direct_course
                    and is_visible
                ):
                    self.direct_course.extended_object.title_set.update(
                        publisher_state=PUBLISHER_STATE_DIRTY
                    )  # mark target page dirty in all languages
                    page = public_instance.direct_course.draft_extension.extended_object
                    page.title_set.update(
                        publisher_state=PUBLISHER_STATE_DIRTY
                    )  # mark source page dirty in all languages
                    break
            elif (
                field.editable
                and not field.auto_created
                and getattr(public_instance, field.name) != getattr(self, field.name)
                and is_visible
            ):
                self.direct_course.extended_object.title_set.update(
                    publisher_state=PUBLISHER_STATE_DIRTY
                )  # mark page dirty in all languages
                break

    def save(self, *args, **kwargs):
        """Enforce validation each time an instance is saved."""
        self.full_clean()
        super().save(*args, **kwargs)

    # pylint: disable=signature-differs
    def delete(self, *args, **kwargs):
        """
        Mark the related course page as dirty if the course about to be deleted was
        published and visible (not to be scheduled).
        """
        try:
            # pylint: disable=no-member
            public_course_run = self.public_course_run
        except CourseRun.DoesNotExist:
            pass
        else:
            if public_course_run.state["priority"] < CourseState.TO_BE_SCHEDULED:
                self.direct_course.extended_object.title_set.update(
                    publisher_state=PUBLISHER_STATE_DIRTY
                )  # mark page dirty in all languages
        return super().delete(*args, **kwargs)

    # pylint: disable=too-many-return-statements
    @staticmethod
    def compute_state(start, end, enrollment_start, enrollment_end):
        """
        Compute at the current time the state of a course run that would have the dates
        passed in argument.

        A static method not using the instance allows to call it with an Elasticsearch result.
        """
        if not start or not enrollment_start:
            return CourseState(CourseState.TO_BE_SCHEDULED)

        # course run end dates are not required and should default to forever
        # e.g. a course run with no end date is presumed to be always on-going
        end = end or MAX_DATE
        enrollment_end = enrollment_end or MAX_DATE

        now = timezone.now()
        if start < now:
            if end > now:
                if enrollment_end > now:
                    # ongoing open
                    return CourseState(CourseState.ONGOING_OPEN, enrollment_end)
                # ongoing closed
                return CourseState(CourseState.ONGOING_CLOSED)
            if enrollment_start < now < enrollment_end:
                # archived open
                return CourseState(CourseState.ARCHIVED_OPEN, enrollment_end)
            # archived closed
            return CourseState(CourseState.ARCHIVED_CLOSED)
        if enrollment_start > now:
            # future not yet open
            return CourseState(CourseState.FUTURE_NOT_YET_OPEN, start)
        if enrollment_end > now:
            # future open
            return CourseState(CourseState.FUTURE_OPEN, start)
        # future already closed
        return CourseState(CourseState.FUTURE_CLOSED)

    @property
    def state(self):
        """Return the state of the course run at the current time."""
        return self.compute_state(
            self.start, self.end, self.enrollment_start, self.enrollment_end
        )

    def get_course(self):
        """Get the course for this course run."""
        is_draft = self.direct_course.extended_object.publisher_is_draft
        ancestor_nodes = self.direct_course.extended_object.node.get_ancestors()
        return Course.objects.filter(
            # Joining on `cms_pages` generate duplicates for courses that are under a parent page
            # when this page exists both in draft and public versions. We need to exclude the
            # parent public page to avoid this duplication
            Q(
                extended_object__node__cms_pages__publisher_is_draft=is_draft
            )  # course has a parent
            | Q(extended_object__node__isnull=True),  # course has no parent
            # Target courses that are ancestors of the course related to the course run
            Q(id=self.direct_course_id) | Q(extended_object__node__in=ancestor_nodes),
            # Exclude snapshots
            extended_object__node__parent__cms_pages__course__isnull=True,  # exclude snapshots
            # Get the course in the same version as the course run
            extended_object__publisher_is_draft=is_draft,
        ).distinct()[0]


class CourseRunTranslation(TranslatedFieldsModel):
    """
    CourseRun Translation model.

    Django parler model linked to the CourseRun to internationalize the fields.
    """

    master = models.ForeignKey(CourseRun, models.CASCADE, related_name="translations")
    title = models.CharField(_("title"), null=True, blank=True, max_length=255)

    class Meta:
        db_table = "richie_course_run_translation"
        unique_together = ("language_code", "master")
        verbose_name = _("Course run translation")
        verbose_name_plural = _("Course run translations")

    def __str__(self):
        """Human representation of a course run translation."""
        return "{model}: {name}".format(
            model=self._meta.verbose_name.title(), name=self.title
        )

    # pylint: disable=signature-differs
    def save(self, *args, **kwargs):
        """
        Mark related course page dirty if the title has changed compared to the public version.
        """
        if (
            # Does the public translation have a different title?
            self.__class__.objects.filter(
                master__draft_course_run__translations__pk=self.pk,
                language_code=self.language_code,
            )
            .exclude(title=self.title)
            .exists()
        ):
            self.master.direct_course.extended_object.title_set.filter(
                language=self.language_code
            ).update(
                publisher_state=PUBLISHER_STATE_DIRTY
            )  # mark page dirty

        return super().save(*args, **kwargs)


class CoursePluginModel(CMSPlugin):
    """
    Course plugin model handles the relation from CoursePlugin
    to their Course instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="course_plugins",
        limit_choices_to={
            # Joining on `cms_pages` generate duplicates for courses that are under a parent page
            # when this page exists both in draft and public versions. We need to exclude the
            # parent public page to avoid this duplication with the first condition.
            # The second condition makes sure the parent is not a course to exclude snapshots.
            # The third condition filters out public course.
            # The fourth condition makes sure only courses show up.
            "node__parent__cms_pages__publisher_is_draft": True,  # exclude course published parent
            "node__parent__cms_pages__course__isnull": True,  # limit to course - no snapshot
            "publisher_is_draft": True,  # plugins work with draft instances
            "course__isnull": False,  # limit to pages linked to a course object
        },
    )
    variant = models.CharField(
        _("variant"),
        max_length=50,
        choices=defaults.COURSE_GLIMPSE_VARIANT_CHOICES,
        help_text=_("Optional glimpse variant for a custom look."),
        blank=True,
        null=True,
    )

    class Meta:
        db_table = "richie_course_plugin"
        verbose_name = _("course plugin")
        verbose_name_plural = _("course plugins")

    def __str__(self):
        """Human representation of a page plugin"""
        return self.page.get_title()


class Licence(TranslatableModel):
    """
    Licence model.

    Instances of this models should only be created by administrators.
    """

    name = TranslatedField()
    content = TranslatedField()
    logo = FilerImageField(
        verbose_name=_("logo"), on_delete=models.PROTECT, related_name="licence"
    )
    url = models.CharField(_("url"), blank=True, max_length=255)
    # Deprecated non-translated fields for name & content
    # Kept around to avoid a breaking change wrt. blue-green deployments
    name_deprecated = models.CharField(_("name"), db_column="name", max_length=200)
    content_deprecated = models.TextField(
        _("content"), blank=False, db_column="content", default=""
    )

    class Meta:
        db_table = "richie_licence"
        verbose_name = _("licence")
        verbose_name_plural = _("licences")

    def __str__(self):
        """Human representation of a licence."""
        return "{model}: {name}".format(
            model=self._meta.verbose_name.title(), name=self.name
        )


class LicenceTranslation(TranslatedFieldsModel):
    """
    Licence Translation model.

    Django parler model linked to the Licence to internationalize the fields.
    """

    master = models.ForeignKey(Licence, models.CASCADE, related_name="translations")
    name = models.CharField(_("name"), max_length=200)
    content = models.TextField(_("content"), blank=False, default="")

    class Meta:
        db_table = "richie_licence_translation"
        unique_together = ("language_code", "master")
        verbose_name = _("Licence translation")
        verbose_name_plural = _("licence translations")

    def __str__(self):
        """Human representation of a licence translation."""
        return "{model}: {name}".format(
            model=self._meta.verbose_name.title(), name=self.name
        )


class LicencePluginModel(CMSPlugin):
    """
    Licence plugin model.
    """

    licence = models.ForeignKey(Licence, on_delete=models.CASCADE)
    description = models.TextField(_("description"), blank=True, default="")

    class Meta:
        db_table = "richie_licence_plugin"
        verbose_name = _("licence plugin")
        verbose_name_plural = _("licence plugins")

    def __str__(self):
        """Human representation of a licence plugin."""
        return "{model:s}: {name:s}".format(
            model=self._meta.verbose_name.title(), name=self.licence.name
        )


extension_pool.register(Course)



"""
API routes exposed by our Courses app.
"""
from django.urls import re_path

from rest_framework import routers

from .api import CourseRunsViewSet, course_runs_sync

ROUTER = routers.SimpleRouter()

# Define our app's API routes with the router
ROUTER.register("course-runs", CourseRunsViewSet, "course_runs")

urlpatterns = ROUTER.urls + [
    re_path("course-runs-sync/?$", course_runs_sync, name="course_run_sync"),
]



"""
Declare and configure the models for the courses application
"""
from django.conf import settings
from django.db import models
from django.utils.translation import gettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models.pluginmodel import CMSPlugin

from ...core.models import BasePageExtension
from ..defaults import CATEGORIES_PAGE, CATEGORY_GLIMPSE_VARIANT_CHOICES


class Category(BasePageExtension):
    """
    The category page extension represents and records a thematic in the catalog.

    This model should be used to record structured data about the thematic whereas the
    associated page object is where we record the less structured information to display on the
    page that presents the thematic.
    """

    color = models.CharField(max_length=10, blank=True, null=True)

    PAGE = CATEGORIES_PAGE

    class Meta:
        db_table = "richie_category"
        ordering = ["-pk"]
        verbose_name = _("category")
        verbose_name_plural = _("categories")

    def __str__(self):
        """Human representation of a category"""
        ancestors = self.extended_object.get_ancestor_pages().filter(
            category__isnull=False
        )
        return "{ancestors:s}{title:s}".format(
            ancestors="{:s} / ".format(" / ".join([a.get_title() for a in ancestors]))
            if ancestors
            else "",
            title=self.extended_object.get_title(),
        )

    def get_es_id(self):
        """
        An ID built with the node path and the position of this category in the taxonomy:
            - P: parent, the category has children
            - L: leaf, the category has no children

        For example: a parent category `P_00010002` and its child `L_000100020001`.
        """
        page = self.extended_object
        return f"{'L' if page.node.is_leaf() else 'P':s}-{page.node.path:s}"

    def get_meta_category(self):
        """
        Get the meta category the current category falls under. Meta-categories are a special kind
        of categories that are not linked to courses but help us organize categories into
        different kinds that are handled separately.
        """
        # Shorcut to the category's page node
        node = self.extended_object.node

        self_is_draft = self.extended_object.publisher_is_draft
        return Category.objects.get(
            extended_object__publisher_is_draft=self_is_draft,
            extended_object__node__path__in=[
                node.path[0:pos] for pos in range(0, len(node.path), node.steplen)[1:]
            ],
            extended_object__node__parent__cms_pages__category__isnull=True,
            # Avoid duplicates from the join on cms_pages
            extended_object__node__parent__cms_pages__publisher_is_draft=self_is_draft,
        )

    def get_courses(self, language=None, include_descendants=True):
        """
        Return a query to get the courses related to this category ie for which a plugin for
        this category is linked to the course page via any placeholder.
        """
        return self.get_reverse_related_page_extensions(
            "course", language=language, include_descendants=include_descendants
        ).filter(extended_object__node__parent__cms_pages__course__isnull=True)

    def get_blogposts(self, language=None, include_descendants=True):
        """
        Return a query to get the blogposts related to this category ie for which a
        plugin for this category is linked to the blogpost page via any placeholder.
        """
        return self.get_reverse_related_page_extensions(
            "blogpost", language=language, include_descendants=include_descendants
        )

    def get_organizations(self, language=None, include_descendants=True):
        """
        Return a query to get the organizations related to this category ie for which a plugin
        for this category is linked to the organization page via any placeholder.
        """
        return self.get_reverse_related_page_extensions(
            "organization", language=language, include_descendants=include_descendants
        )

    def get_persons(self, language=None, include_descendants=True):
        """
        Return a query to get the persons related to this category ie for which a plugin for
        this category is linked to the person page via any placeholder.
        """
        return self.get_reverse_related_page_extensions(
            "person", language=language, include_descendants=include_descendants
        )

    def get_children_categories(self):
        """
        Return a query to get only the children pages that are Category
        objects related to this category. A convenient method since
        ``page.get_child_pages`` return every direct child page, no matter
        it's a Category or not.
        """
        return Category.objects.filter(
            extended_object__in=self.extended_object.get_child_pages()
        ).select_related("extended_object")


def get_category_limit_choices_to():
    """Return a query limiting the categories proposed when creating a CategoryPlugin."""
    limit_choices_to = {
        # Joining on `cms_pages` generate duplicates for categories that are under a parent page
        # when this page exists both in draft and public versions. We need to exclude the
        # parent public page to avoid this duplication with the first condition.
        # The second condition makes sure the parent is a category.
        # The third condition filters out public category.
        # The fourth condition makes sure only categories show up.
        "node__parent__cms_pages__publisher_is_draft": True,  # exclude category published parent
        "node__parent__cms_pages__category__isnull": False,  # exclude meta categories
        "publisher_is_draft": True,  # plugins work with draft instances
        "category__isnull": False,  # limit to pages linked to a category object
    }
    # Limit to leaf categories only if active in settings (False by default)
    if getattr(settings, "LIMIT_PLUGIN_CATEGORIES_TO_LEAF", False):
        limit_choices_to["node__numchild"] = 0

    return limit_choices_to


class CategoryPluginModel(CMSPlugin):
    """
    Category plugin model handles the relation between CategoryPlugin
    and their Category instance

    We only propose to link category objects:
    - are not at the root of the category tree (this position defines a meta category which
      is used to define a filter bank on the search page e.g. Subjects, Levels).
    - that are a leaf in the category tree (can be deactivated with setting
      `LIMIT_PLUGIN_CATEGORIES_TO_LEAF`),
    """

    page = models.ForeignKey(
        Page,
        related_name="category_plugins",
        limit_choices_to=get_category_limit_choices_to,
        on_delete=models.CASCADE,
    )
    variant = models.CharField(
        _("variant"),
        max_length=50,
        choices=CATEGORY_GLIMPSE_VARIANT_CHOICES,
        help_text=_("Optional glimpse variant for a custom look."),
        blank=True,
        null=True,
    )

    class Meta:
        db_table = "richie_category_plugin"
        verbose_name = _("category plugin")
        verbose_name_plural = _("category plugins")

    def __str__(self):
        """Human representation of a page plugin"""
        return self.page.get_title()


extension_pool.register(Category)



# Generated by Django 2.1.7 on 2019-02-23 01:18

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import filer.fields.image


class Migration(migrations.Migration):

    dependencies = [("large_banner", "0002_add_template_and_content_fields")]

    operations = [
        migrations.AlterField(
            model_name="largebanner",
            name="logo",
            field=filer.fields.image.FilerImageField(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.PROTECT,
                related_name="logo",
                to=settings.FILER_IMAGE_MODEL,
                verbose_name="logo",
            ),
        ),
        migrations.AlterField(
            model_name="largebanner",
            name="logo_alt_text",
            field=models.CharField(blank=True, max_length=255),
        ),
    ]



"""
Toolbar extension for the courses application
"""
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _

from cms.api import get_page_draft
from cms.extensions.toolbar import ExtensionToolbar
from cms.toolbar_pool import toolbar_pool
from cms.utils.page_permissions import user_can_add_subpage, user_can_change_page
from cms.utils.urlutils import admin_reverse

from .defaults import PAGE_EXTENSION_TOOLBAR_ITEM_POSITION
from .models import Category, Course, Organization, Person


class BaseExtensionToolbar(ExtensionToolbar):
    """
    This extension class customizes the toolbar for our page extensions.

    It must be inherited to define the `model` class attribute.
    """

    def populate(self):
        """
        When visiting a page that is linked to a page extension, show an item in the toolbar only
        to users who have the permission to update the page.

        Clicking on this menu item will open the admin page of the page extension.

        Note: We don't want to show a link to create a new page extension on a page that does not
        have it yet. For clarity, we choose to only create page extensions transparently via the
        page creation wizard.
        """
        # always use draft if we have a page
        self.page = get_page_draft(self.request.current_page)
        if not self.page:
            # Nothing to do
            return

        if user_can_change_page(user=self.request.user, page=self.page):
            # Add page extension edition menu item
            page_extension, admin_url = self.get_page_extension_admin()
            if page_extension:
                edit_mode_inactive = not self.toolbar.edit_mode_active
                page_menu = self.toolbar.get_or_create_menu("page")
                # Create the new menu item as a modal
                page_menu.add_modal_item(
                    _("{!s} settings").format(
                        capfirst(str(self.model._meta.verbose_name))
                    ),
                    url=admin_url,
                    disabled=edit_mode_inactive,
                    position=PAGE_EXTENSION_TOOLBAR_ITEM_POSITION,
                )


@toolbar_pool.register
class CategoryExtensionToolbar(BaseExtensionToolbar):
    """
    This extension class customizes the toolbar for the category page extension
    """

    model = Category


@toolbar_pool.register
class CourseExtensionToolbar(BaseExtensionToolbar):
    """
    This extension class customizes the toolbar for the course page extension
    """

    model = Course

    def populate(self):

        super().populate()
        if not self.page:
            # Nothing to do
            return

        if user_can_change_page(
            user=self.request.user, page=self.page
        ) and user_can_add_subpage(self.request.user, self.page):
            # Add snapshot menu item only on course pages
            try:
                course = Course.objects.only("id").get(extended_object=self.page)
            except Course.DoesNotExist:
                return

            # Course snapshots can not be snapshotted
            if self.page.parent_page:
                try:
                    self.page.parent_page.course
                except Course.DoesNotExist:
                    pass
                else:
                    return

            url = admin_reverse("cms_course_snapshot", kwargs={"course_id": course.id})
            edit_mode_inactive = not self.toolbar.edit_mode_active
            page_menu = self.toolbar.get_or_create_menu("page")
            # Create the new menu item as an Ajax call
            page_menu.add_ajax_item(
                _("Snapshot this page..."),
                action=url,
                data={},
                question=_(
                    "This will place a copy of this page as its child and move all its course"
                    "runs as children of its new copy."
                ),
                on_success=self.toolbar.REFRESH_PAGE,
                disabled=edit_mode_inactive,
                position=PAGE_EXTENSION_TOOLBAR_ITEM_POSITION - 1,
            )


@toolbar_pool.register
class OrganizationExtensionToolbar(BaseExtensionToolbar):
    """
    This extension class customizes the toolbar for the organization page extension
    """

    model = Organization


@toolbar_pool.register
class PersonExtensionToolbar(BaseExtensionToolbar):
    """
    This extension class customizes the toolbar for the person page extension
    """

    model = Person



"""
API endpoints for the courses app.
"""
import hashlib
import hmac

from django.conf import settings
from django.db.models import Q

from cms import signals as cms_signals
from cms.models import Page
from rest_framework.decorators import api_view
from rest_framework.permissions import BasePermission
from rest_framework.response import Response
from rest_framework.viewsets import ModelViewSet

from .lms import LMSHandler
from .models import Course, CourseRun, CourseRunSyncMode
from .serializers import CourseRunSerializer


class NotAllowed(BasePermission):
    """
    Utility permission class to deny all requests. This is used as a default to close
    requests to unsupported actions.
    """

    def has_permission(self, request, view):
        """
        Always deny permission.
        """
        return False


class CourseRunsViewSet(ModelViewSet):
    """
    API endpoints to access and perform actions on course runs.
    """

    permission_classes = [NotAllowed]
    queryset = CourseRun.objects.all()
    serializer_class = CourseRunSerializer

    def get_permissions(self):
        """
        Manage permissions for builtin DRF methods on ViewSets.
        """
        if self.action == "retrieve":
            permission_classes = []
        else:
            try:
                permission_classes = getattr(self, self.action).kwargs.get(
                    "permission_classes"
                )
            except AttributeError:
                permission_classes = self.permission_classes

        return [permission() for permission in permission_classes]


# pylint: disable=too-many-return-statements,unused-argument, too-many-locals,too-many-branches
@api_view(["POST"])
def course_runs_sync(request, version):
    """View for the web hook to create or update course runs based on their resource link.

    - An existing course run is updated only if its "sync_mode" field is set to something else
      than "manual".

    - The public version of a course run is updated only if its "sync_mode" field is set to
      "sync_to_public". Otherwise, only the draft version is updated and the related course
      is marked dirty.

    - A new course run is created only if the "DEFAULT_COURSE_RUN_SYNC_MODE" parameter is set
      to something else than "manual" in the lms configuration (or the setting
      "RICHIE_DEFAULT_COURSE_RUN_SYNC_MODE" in the absence of LMS preference). Otherwise, only
      existing course runs are updated.

    - A new public course run is created only if the "DEFAULT_COURSE_RUN_SYNC_MODE" parameter
      is set to "sync_to_public" in the lms configuration (or the setting
      "RICHIE_DEFAULT_COURSE_RUN_SYNC_MODE" in the absence of LMS preference). Otherwise, only
      the draft course run is created and the related course is marked dirty.

    Parameters
    ----------
    request : Type[django.http.request.HttpRequest]
        The request on the API endpoint, it should contain a payload with course run fields.

    Returns
    -------
    Type[rest_framework.response.Response]
        HttpResponse acknowledging the success or failure of the synchronization operation.
    """
    msg = request.body.decode("utf-8")

    # Check if the provided signature is valid against any secret in our list
    #
    # We need to do this to support 2 or more versions of our infrastructure at the same time.
    # It then enables us to do updates and change the secret without incurring downtime.
    authorization_header = request.headers.get("Authorization")
    if not authorization_header:
        return Response("Missing authentication.", status=401)

    signature_is_valid = any(
        authorization_header
        == "SIG-HMAC-SHA256 {:s}".format(
            hmac.new(
                secret.encode("utf-8"),
                msg=msg.encode("utf-8"),
                digestmod=hashlib.sha256,
            ).hexdigest()
        )
        for secret in getattr(settings, "RICHIE_COURSE_RUN_SYNC_SECRETS", [])
    )

    if not signature_is_valid:
        return Response("Invalid authentication.", status=403)

    # Select LMS from resource link
    resource_link = request.data.get("resource_link")
    if not resource_link:
        return Response({"resource_link": ["This field is required."]}, status=400)

    lms = LMSHandler.select_lms(resource_link)
    if lms is None:
        return Response(
            {"resource_link": ["No LMS configuration found for this resource link."]},
            status=400,
        )
    sync_mode = lms.default_course_run_sync_mode

    target_course_runs = CourseRun.objects.filter(resource_link=resource_link)
    draft_course_runs = target_course_runs.filter(draft_course_run__isnull=True)

    serializer = lms.get_course_run_serializer(
        request.data, partial=bool(draft_course_runs)
    )

    if serializer.is_valid() is not True:
        return Response(serializer.errors, status=400)

    if draft_course_runs:
        # Remove protected fields before update
        cleaned_data = lms.clean_course_run_data(serializer.validated_data)

        for course_run in draft_course_runs.filter(
            sync_mode__in=[
                CourseRunSyncMode.SYNC_TO_DRAFT,
                CourseRunSyncMode.SYNC_TO_PUBLIC,
            ]
        ):
            nb_updated = CourseRun.objects.filter(
                Q(pk=course_run.pk)
                | Q(
                    draft_course_run__sync_mode=CourseRunSyncMode.SYNC_TO_PUBLIC,
                    draft_course_run=course_run,
                )
            ).update(**cleaned_data)

            public_course = course_run.direct_course.public_extension
            if course_run.sync_mode == CourseRunSyncMode.SYNC_TO_PUBLIC:
                if public_course:
                    # If the public course run did not exist yet it has to be created
                    if nb_updated == 1:
                        public_course.copy_relations(course_run.direct_course)

                    # What we did has changed the public course page. We must reindex it
                    cms_signals.post_publish.send(
                        sender=Page,
                        instance=course_run.direct_course.extended_object,
                        language=None,
                    )
            else:
                course_run.refresh_from_db()
                course_run.mark_course_dirty()
        return Response({"success": True})

    # We need to create a new course run
    if lms.default_course_run_sync_mode == CourseRunSyncMode.MANUAL:
        return Response(
            {"resource_link": ["Unknown course run when creation is deactivated."]},
            status=400,
        )

    # Look for the course targeted by the resource link
    course_number = lms.extract_course_number(request.data)
    try:
        course = Course.objects.get(
            code=course_number, extended_object__publisher_is_draft=True
        )
    except Course.DoesNotExist:
        return Response(
            {"resource_link": [f"Unknown course: {course_number:s}."]}, status=400
        )

    # Instantiate a new draft course run
    draft_course_run = CourseRun(
        direct_course=course, sync_mode=sync_mode, **serializer.validated_data
    )

    # Create the related public course run if necessary
    if sync_mode == CourseRunSyncMode.SYNC_TO_PUBLIC:
        # Don't mark the related course page dirty and directly add
        # the course run to the corresponding public course page
        draft_course_run.save()
        if course.public_extension_id:
            public_course_run = CourseRun(
                direct_course=course.public_extension,
                draft_course_run=draft_course_run,
                sync_mode=sync_mode,
                **serializer.validated_data,
            )
            public_course_run.save()

            # What we did has changed the public course page. We must reindex it
            cms_signals.post_publish.send(
                sender=Page, instance=course.extended_object, language=None
            )
    else:
        # Save the draft course run marking the course page dirty
        draft_course_run.save()
        draft_course_run.mark_course_dirty()

    return Response({"success": target_course_runs.exists()})



"""
Custom widgets for the course application
"""
from django.contrib.admin.widgets import AdminSplitDateTime
from django.utils import timezone


class CourseRunSplitDateTimeWidget(AdminSplitDateTime):
    """
    A SplitDateTime Widget that displays a default time. We can't use a Django initial value on
    the form because it would set both the date and the time.

    This widget must be combined with the "CourseRunSplitDateTimeField" so that the form knows
    how to ignore our default time if it is submitted without a date.
    """

    def get_context(self, name, value, attrs):
        """
        Fill the value of the time field with our default time.
        """
        context = super().get_context(name, value, attrs)
        if context["widget"]["subwidgets"][1]["value"] is None:
            context["widget"]["subwidgets"][1]["value"] = (
                timezone.now().replace(hour=0, minute=0, second=0, microsecond=0).time()
            )
        return context



# Generated by Django 3.1.5 on 2021-01-26 21:00

from django.db import migrations


def delete_orphan_course_run_nodes(apps, schema_editor):
    """
    When we deleted the pages previously related to course runs, we
    forgot to delete the related nodes. Let's fix this.
    """
    TreeNode = apps.get_model("cms", "TreeNode")

    for node in (
        TreeNode.objects.filter(
            cms_pages__isnull=True, parent__cms_pages__course__isnull=False
        )
        .distinct()
        .iterator()
    ):
        # When deleting a node, we must also update its parent to decrement
        # its number of children
        parentpath = node.path[0 : (node.depth - 1) * 4]
        parent = TreeNode.objects.get(path=parentpath)
        node.delete()
        parent.numchild -= 1
        parent.save()


class Migration(migrations.Migration):

    dependencies = [("courses", "0027_auto_20210119_1411")]

    operations = [
        migrations.RunPython(delete_orphan_course_run_nodes, migrations.RunPython.noop)
    ]



# Generated by Django 2.1.7 on 2019-02-22 01:57

from django.db import migrations, models

from ..defaults import SECTION_TEMPLATES


class Migration(migrations.Migration):

    dependencies = [("section", "0001_initial")]

    operations = [
        migrations.AddField(
            model_name="section",
            name="template",
            field=models.CharField(
                choices=SECTION_TEMPLATES,
                default=SECTION_TEMPLATES[0][0],
                help_text="Optional template for custom look.",
                max_length=150,
                verbose_name="Template",
            ),
        )
    ]



"""
Glimpse CMS plugin factories
"""
import factory

from .defaults import GLIMPSE_VARIANTS
from .models import Glimpse


class GlimpseFactory(factory.django.DjangoModelFactory):
    """
    Factory to create random instances of Glimpse for testing.
    """

    class Meta:
        model = Glimpse

    title = factory.Faker("sentence", nb_words=6)
    variant = GLIMPSE_VARIANTS[0][0]
    image = factory.SubFactory("richie.apps.core.factories.FilerImageFactory")
    content = factory.Faker("text", max_nb_chars=200)



"""
Utility django form field that lets us validate and clean datetimeranges received from the API,
ensuring there is at least one date and any supplied date is valid
"""
import json

from django import forms
from django.core.exceptions import ValidationError

import arrow


class DatetimeRangeField(forms.Field):
    """
    Check each of the datetimess using the basic DateTimeField but with custom date formats to
    support actual ISO strings, and return either None or a tuple as a daterange

    Valid input examples:
    - '["2018-01-01T06:00:00Z", null]'
    - '[null, "2018-01-01T06:00:00Z"]'
    - '["2018-01-01T06:00:00Z", "2018-01-31T06:00:00Z"]'
    - '["2018-01-01T06:00:00Z", "2018-02-01T06:00:00+04:00"]'
    - '["2018-01-01T06:00:00", "2018-02-01T06:00:00Z"]'
    """

    def clean(self, value):
        """
        Validate our string and each of the datetimes passed into it, using regular DateTimeFields
        for item validation
        """
        # Handle null values: either we're missing a required value
        if not value and self.required:
            raise ValidationError("Missing required field")
        # Or the null value is optional: return None and exit
        if not value:
            return None

        # Decode our incoming JSON array and unpack it all at once
        try:
            iso_start, iso_end = json.loads(value)
        # JSON parsing failed: blame the formatting
        except json.JSONDecodeError as error:
            raise ValidationError("Invalid JSON formatting") from error
        # Unpacking failed: input was not a list of a least 2 values
        except ValueError as error:
            raise ValidationError("Empty datetimerange is invalid") from error
        # Sanity check: don't throw a 500 when the param is entirely wrong
        except TypeError as error:
            raise ValidationError(
                "Invalid parameter type: must be a JSON array"
            ) from error

        # Reject input that is made up of falsy values
        if not iso_start and not iso_end:
            raise ValidationError("A valid datetimerange needs at least 1 datetime")

        try:
            datetime_range = tuple(
                arrow.get(iso_datetime) if iso_datetime else None
                for iso_datetime in (iso_start, iso_end)
            )
        except arrow.parser.ParserError as error:
            raise ValidationError("Invalid datetime format; use ISO 8601") from error

        return datetime_range



"""
Base backend to connect richie with an LMS
"""
from django.conf import settings

from ..models.course import CourseRunSyncMode


class BaseLMSBackend:
    """
    Base backend to hold the methods common to all backends and provide a skeleton for others.
    """

    def __init__(self, configuration, *args, **kwargs):
        """Attach configuration to the backend instance."""
        super().__init__(*args, **kwargs)
        self.configuration = configuration

    @property
    def default_course_run_sync_mode(self):
        """Course run synchronization mode."""
        return self.configuration.get("DEFAULT_COURSE_RUN_SYNC_MODE") or getattr(
            settings,
            "RICHIE_DEFAULT_COURSE_RUN_SYNC_MODE",
            CourseRunSyncMode.SYNC_TO_PUBLIC,
        )



# Generated by Django 2.2.15 on 2020-08-27 08:11

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0016_auto_20200417_1237"),
    ]

    operations = [
        migrations.AlterField(
            model_name="courserun",
            name="resource_link",
            field=models.CharField(
                blank=True, max_length=200, null=True, verbose_name="Resource link"
            ),
        ),
    ]



"""Define SearchConfig as the default app configuration."""
from django.conf import settings

from elasticsearch import Elasticsearch

# pylint: disable=invalid-name
default_app_config = "richie.apps.search.apps.SearchConfig"

ES_CLIENT = Elasticsearch([getattr(settings, "RICHIE_ES_HOST", "elasticsearch")])



"""
Specific exceptions for the search app
"""


class ApiConsumingException(Exception):
    """
    Exception raised when an exception occurs during API walk using our api_consumption util
    """


class IndexerDataException(Exception):
    """
    Exception raised when an Indexer fails to provide the data required for ES indexing
    """


class QueryFormatException(Exception):
    """
    Exception raised when the query parameters for a GET list call are invalid
    """



"""
NestedItem plugin default settings
"""
from django.conf import settings
from django.utils.translation import gettext_lazy as _

ACCORDION, LIST = "accordion", "list"

NESTEDITEM_VARIANTS = getattr(
    settings,
    "RICHIE_NESTEDITEM_VARIANTS",
    [(LIST, _("List")), (ACCORDION, _("Accordion"))],
)



"""Django Configuration mixins for the Richie courses app."""
from configurations.utils import uppercase_attributes

from .. import settings


def apply_settings(cls):
    """A decorator to set class attributes from imported settings."""
    for key, value in uppercase_attributes(settings).items():
        setattr(cls, key, value)
    return cls


@apply_settings
class RichieCoursesConfigurationMixin:
    """
    A Django Configuration mixin to set sensible defaults for the settings of Richie's
    courses app.
    """



"""
Glimpse plugin models
"""
from django.db import models
from django.utils.text import Truncator
from django.utils.translation import gettext_lazy as _

from cms.models.fields import PageField
from cms.models.pluginmodel import CMSPlugin
from filer.fields.image import FilerImageField

from .defaults import GLIMPSE_VARIANTS


class Glimpse(CMSPlugin):
    """
    Glimpse plugin model
    """

    title = models.CharField(max_length=255, blank=True, null=True)
    variant = models.CharField(
        _("Variant"),
        max_length=50,
        choices=GLIMPSE_VARIANTS,
        default=GLIMPSE_VARIANTS[0][0],
        blank=True,
        null=True,
        help_text=_("Form factor variant"),
    )
    image = FilerImageField(
        related_name="image",
        verbose_name=_("image"),
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )
    content = models.TextField(_("Content"), null=True, blank=True, default="")
    link_url = models.URLField(
        verbose_name=_("External URL"),
        blank=True,
        null=True,
        max_length=255,
        help_text=_("Make the glimpse as a link with an external URL."),
    )
    link_page = PageField(
        verbose_name=_("Internal URL"),
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        help_text=_("Make the glimpse as a link with an internal (page) URL."),
    )

    def __str__(self):
        return Truncator(self.title).words(6, truncate="...")

    def get_link(self):
        """
        Return attached link if any.

        In case the "link_url" and "link_page" fields are both not empty,
        "link_url" is always used.
        """
        link = None

        if self.link_url:
            link = self.link_url
        elif self.link_page_id:
            # pylint: disable=E1101
            link = self.link_page.get_absolute_url(language=self.language)

        return link

    def is_blank_target(self):
        """
        Return if the possible link is to be open in a blank target
        (new window/tab) or not.

        Blank target is only true if object has a value for "link_url"
        attribute.
        """
        return bool(self.link_url)



"""
Declare and configure the models for the courses application
"""
from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.text import slugify
from django.utils.translation import gettext_lazy as _

from cms.api import Page, PagePermission
from cms.extensions.extension_pool import extension_pool
from cms.models import CMSPlugin
from filer.models import FolderPermission

from ...core.helpers import get_permissions
from ...core.models import BasePageExtension
from .. import defaults
from .role import PageRole


class Organization(BasePageExtension):
    """
    The organization page extension represents and records entities that manage courses.
    It could be a university or a training company for example.

    This model should be used to record structured data about the organization whereas the
    associated page object is where we record the less structured information to display on the
    page to present the organization.
    """

    code = models.CharField(
        _("code"), db_index=True, max_length=100, null=True, blank=True
    )

    PAGE = defaults.ORGANIZATIONS_PAGE

    class Meta:
        db_table = "richie_organization"
        verbose_name = _("organization")
        verbose_name_plural = _("organizations")
        ordering = ["-pk"]

    def __str__(self):
        """Human representation of an organization"""
        return "{model}: {name} ({code})".format(
            code=self.code,
            name=self.extended_object.get_title(),
            model=self._meta.verbose_name.title(),
        )

    def get_es_id(self):
        """
        An ID built with the node path and the position of this organization in the taxonomy:
            - P: parent, the organization has children
            - L: leaf, the organization has no children

        For example: a parent organization `P_00010002` and its child `L_000100020001`.
        """
        page = self.extended_object
        return f"{'L' if page.node.is_leaf() else 'P':s}-{page.node.path:s}"

    def clean(self):
        """
        We normalize the code with slugify for better uniqueness
        """
        if self.code:
            # Normalize the code by slugifying and capitalizing it
            self.code = slugify(self.code, allow_unicode=True).upper()
        return super().clean()

    def validate_unique(self, exclude=None):
        """
        We can't rely on a database constraint for uniqueness because pages
        exist in two versions: draft and published.
        """
        if self.code:
            # Check unicity for the version being saved (draft or published)
            is_draft = self.extended_object.publisher_is_draft
            uniqueness_query = self.__class__.objects.filter(
                code=self.code, extended_object__publisher_is_draft=is_draft
            )

            # If the page is being updated, we should exclude it while looking for duplicates
            if self.pk:
                uniqueness_query = uniqueness_query.exclude(pk=self.pk)

            # Raise a ValidationError if the code already exists
            if uniqueness_query.exists():
                raise ValidationError(
                    {"code": ["An Organization already exists with this code."]}
                )
        return super().validate_unique(exclude=exclude)

    def create_page_role(self):
        """
        Create a new page role for the organization with:
          - a user group to handle permissions for admins of this organization,
          - a folder in Django Filer to store images related to this organization,
          - all necessary permissions.
        """
        if not getattr(settings, "CMS_PERMISSION", False):
            return None

        # The page role is only created for draft organizations
        if not self.extended_object.publisher_is_draft:
            return None

        # Don't do anything if it already exists
        page_role = PageRole.objects.filter(
            page=self.extended_object, role=defaults.ADMIN
        ).first()

        if page_role:
            return page_role

        # Create a role for admins of this organization (which will automatically create a new
        # user group and a new Filer folder)
        page_role = PageRole.objects.create(
            page=self.extended_object, role=defaults.ADMIN
        )

        # Associate permissions as defined in settings:
        # - Create Django permissions
        page_role.group.permissions.set(
            get_permissions(
                defaults.ORGANIZATION_ADMIN_ROLE.get("django_permissions", [])
            )
        )

        # - Create DjangoCMS page permissions
        PagePermission.objects.create(
            group_id=page_role.group_id,
            page=self.extended_object,
            **defaults.ORGANIZATION_ADMIN_ROLE.get("organization_page_permissions", {}),
        )

        # - Create the Django Filer folder permissions
        FolderPermission.objects.create(
            folder_id=page_role.folder_id,
            group_id=page_role.group_id,
            **defaults.ORGANIZATION_ADMIN_ROLE.get(
                "organization_folder_permissions", {}
            ),
        )
        return page_role

    def save(self, *args, **kwargs):
        """
        Enforce validation on each instance save
        """
        self.full_clean()
        super().save(*args, **kwargs)

    def get_courses(self, language=None):
        """
        Return a query to get the courses related to this organization ie for which a plugin for
        this organization is linked to the course page via any placeholder.
        """
        return self.get_reverse_related_page_extensions(
            "course", language=language
        ).filter(extended_object__node__parent__cms_pages__course__isnull=True)

    def get_persons(self, language=None):
        """
        Return a query to get the persons related to this organization ie for which a plugin for
        this organization is linked to the person page via any placeholder.
        """
        return self.get_reverse_related_page_extensions("person", language=language)


class OrganizationPluginModel(CMSPlugin):
    """
    Organization plugin model handles the relation from OrganizationPlugin
    to their Organization instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="organization_plugins",
        limit_choices_to={
            "publisher_is_draft": True,  # plugins work with draft instances
            "organization__isnull": False,  # limit to pages linked to an organization object
        },
    )
    variant = models.CharField(
        _("variant"),
        max_length=50,
        choices=defaults.ORGANIZATION_GLIMPSE_VARIANT_CHOICES,
        help_text=_("Optional glimpse variant for a custom look."),
        blank=True,
        null=True,
    )

    class Meta:
        db_table = "richie_organization_plugin"
        verbose_name = _("organization plugin")
        verbose_name_plural = _("organization plugins")

    def __str__(self):
        """Human representation of a page plugin"""
        return self.page.get_title()


class OrganizationsByCategoryPluginModel(CMSPlugin):
    """
    Handle the relation between a OrganizationsByCategoryPlugin plugin and its Category
    instance.
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="organizations_by_category_plugins",
        limit_choices_to={
            "publisher_is_draft": True,  # plugins work with draft instances
            "category__isnull": False,  # limit to pages linked to a category object
        },
    )
    variant = models.CharField(
        _("variant"),
        max_length=50,
        choices=defaults.ORGANIZATION_GLIMPSE_VARIANT_CHOICES,
        help_text=_("Optional glimpse variant for a custom look."),
        blank=True,
        null=True,
    )

    class Meta:
        db_table = "richie_organizations_by_category_plugin"
        verbose_name = _("organizations by category plugin")
        verbose_name_plural = _("organizations by category plugins")


extension_pool.register(Organization)



# Generated by Django 2.2.8 on 2020-01-02 13:56

from django.conf import settings
from django.core.exceptions import ObjectDoesNotExist
from django.db import migrations


def forwards_func(apps, schema_editor):
    Licence = apps.get_model("courses", "Licence")
    LicenceTranslation = apps.get_model("courses", "LicenceTranslation")

    for licence in Licence.objects.all():
        LicenceTranslation.objects.create(
            master_id=licence.pk,
            language_code=settings.LANGUAGE_CODE,
            name=licence.name_deprecated,
        )


def backwards_func(apps, schema_editor):
    Licence = apps.get_model("courses", "Licence")
    LicenceTranslation = apps.get_model("courses", "LicenceTranslation")

    for licence in Licence.objects.all():
        translation = _get_translation(licence, LicenceTranslation)
        licence.name_deprecated = translation.name
        licence.save()  # Note this only calls Model.save()


def _get_translation(licence, LicenceTranslation):
    translations = LicenceTranslation.objects.filter(master_id=licence.pk)
    try:
        # Try default translation
        return translations.get(language_code=settings.LANGUAGE_CODE)
    except ObjectDoesNotExist:
        try:
            # Try default language
            return translations.get(language_code=settings.PARLER_DEFAULT_LANGUAGE_CODE)
        except ObjectDoesNotExist:
            # Maybe the object was translated only in a specific language?
            # Take the first existing translation
            return translations.first()


class Migration(migrations.Migration):

    dependencies = [("courses", "0012_add_translation_model_for_licence_fields")]

    operations = [migrations.RunPython(forwards_func, backwards_func)]



"""
Declare and configure the models for the program part
"""
from django.db import models
from django.utils.translation import gettext_lazy as _

from cms.api import Page
from cms.extensions.extension_pool import extension_pool
from cms.models.pluginmodel import CMSPlugin

from ...core.models import BasePageExtension
from ..defaults import PROGRAMS_PAGE


class Program(BasePageExtension):
    """
    The program extension represents and records a program.
    """

    PAGE = PROGRAMS_PAGE

    class Meta:
        db_table = "richie_program"
        ordering = ["-pk"]
        verbose_name = _("program")
        verbose_name_plural = _("programs")

    def __str__(self):
        """Human representation of a program"""
        return "{model}: {name}".format(
            model=self._meta.verbose_name.title(), name=self.extended_object.get_title()
        )


class ProgramPluginModel(CMSPlugin):
    """
    Program plugin model handles the relation between the ProgramPluginModel
    and its program instance
    """

    page = models.ForeignKey(
        Page,
        on_delete=models.CASCADE,
        related_name="program_plugins",
        limit_choices_to={"publisher_is_draft": True, "program__isnull": False},
    )

    class Meta:
        db_table = "richie_program_plugin"
        verbose_name = _("program plugin")
        verbose_name_plural = _("program plugins")

    def __str__(self):
        """Human representation of a page plugin"""
        return self.page.get_title()


extension_pool.register(Program)



"""
Section plugin default settings
"""
from django.conf import settings
from django.utils.translation import gettext_lazy as _

SECTION_TEMPLATES = getattr(
    settings,
    "RICHIE_SECTION_TEMPLATES",
    [
        ("richie/section/section.html", _("Default")),
        ("richie/section/section_primary.html", _("Primary scheme")),
        ("richie/section/section_secondary.html", _("Secondary scheme")),
        ("richie/section/section_tertiary.html", _("Tertiary scheme")),
        ("richie/section/section_quaternary.html", _("Quaternary scheme")),
        ("richie/section/section_quinary.html", _("Quinary scheme")),
        ("richie/section/section_senary.html", _("Senary scheme")),
        ("richie/section/section_septenary.html", _("Septenary scheme")),
        ("richie/section/section_tiles.html", _("Item tiles")),
    ],
)



"""
Template context processors
"""
import json

from django.conf import settings
from django.contrib.sites.models import Site
from django.middleware.csrf import get_token

from . import defaults


def site_metas(request):
    """
    Context processor to add all information required by Richie CMS templates and frontend.

    If `CDN_DOMAIN` settings is defined we add it in the context. It allows
    to load statics js on a CDN like cloudfront.
    """
    site_current = Site.objects.get_current()
    protocol = "https" if request.is_secure() else "http"

    context = {
        **{
            f"GLIMPSE_PAGINATION_{k.upper()}": v
            for k, v in {
                **defaults.GLIMPSE_PAGINATION,
                **getattr(settings, "RICHIE_GLIMPSE_PAGINATION", {}),
            }.items()
        },
        "SITE": {
            "name": site_current.name,
            "domain": site_current.domain,
            "web_url": f"{protocol:s}://{site_current.domain:s}",
        },
        "FRONTEND_CONTEXT": {
            "context": {
                "csrftoken": get_token(request),
                "environment": getattr(settings, "ENVIRONMENT", ""),
                "release": getattr(settings, "RELEASE", ""),
                "sentry_dsn": getattr(settings, "SENTRY_DSN", ""),
            }
        },
    }

    if getattr(settings, "CDN_DOMAIN", None):
        context["CDN_DOMAIN"] = settings.CDN_DOMAIN

    authentication_delegation = getattr(
        settings, "RICHIE_AUTHENTICATION_DELEGATION", None
    )
    if authentication_delegation:

        context["AUTHENTICATION"] = {
            "profile_urls": json.dumps(
                {
                    key: {
                        "label": str(url["label"]),
                        "action": str(
                            url["href"].format(
                                base_url=authentication_delegation["BASE_URL"]
                            )
                        ),
                    }
                    for key, url in authentication_delegation.get(
                        "PROFILE_URLS", {}
                    ).items()
                }
            ),
        }

        context["FRONTEND_CONTEXT"]["context"]["authentication"] = {
            "endpoint": authentication_delegation["BASE_URL"],
            "backend": authentication_delegation["BACKEND"],
        }

    if getattr(settings, "RICHIE_LMS_BACKENDS", None):
        context["FRONTEND_CONTEXT"]["context"]["lms_backends"] = [
            {
                "endpoint": lms["BASE_URL"],
                "backend": lms["JS_BACKEND"],
                "course_regexp": lms["JS_COURSE_REGEX"],
            }
            for lms in getattr(settings, "RICHIE_LMS_BACKENDS", [])
        ]

    context["FRONTEND_CONTEXT"] = json.dumps(context["FRONTEND_CONTEXT"])
    return context



"""A multi select field where the array of values is stored as a comma separated string."""
from django.core import checks, exceptions, validators
from django.db import models
from django.forms import MultipleChoiceField, widgets
from django.utils.text import capfirst
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext_lazy


def to_sentence(elements):
    """
    Arguments:
    -----------
    elements (Iterable[string]): an iterable yielding the elements that we want
        to arrange in a sentence.

    Returns:
    --------
    string: the list of elements of the iterable in the form of a sentence:
        ["French"] > "French"
        ["French", "English"] > "French and english"
        ["French", "English", "german"] > "French, english and german"

    """
    if len(elements) > 2:
        return (
            _("{:s} and {:s}").format(
                ", ".join(map(str, elements[:-1])), str(elements[-1])
            )
        ).lower()
    return (_(" and ").join(map(str, elements))).lower()


class MaxChoicesValidator(validators.MaxLengthValidator):
    """A max length validator for the number of choices in a multiselect value."""

    message = _("You can only select up to %(limit_value)d choices.")
    code = "max_choices"


class MultiSelectFormField(MultipleChoiceField):
    """
    The Formfield for our multiselect field is based on the SelectMultiple widget and
    should validate its `max_choices` parameter.
    """

    widget = widgets.SelectMultiple

    def __init__(self, *args, **kwargs):
        """
        Set validators for the field's maximum number of choices.
        """
        self.max_choices = kwargs.pop("max_choices", None)
        super().__init__(*args, **kwargs)

        if self.max_choices is not None:
            self.validators.append(MaxChoicesValidator(self.max_choices))


class MultiSelectField(models.CharField):
    """
    A custom database field to store a list of string values.
    This is an alternative to Django's ArrayField that is compatible with Mysql.
    The array of values is stored in a comma separated string.
    """

    description = _("Multi select field (up to %(max_choices)s choices)")

    error_message = ngettext_lazy(
        "Value %(value)s is not a valid choice.",
        "Values %(value)s are not valid choices.",
        "number",
    )

    def __init__(self, *args, **kwargs):
        """
        Add the `max_choices` parameter and make it required.
        """
        self.max_choices = kwargs.pop("max_choices", None)
        super().__init__(*args, **kwargs)
        self.max_length = kwargs.pop("max_length", None)
        # We only need to validate `max_choices` since its coherence with
        # `max_length` is ensured by below's field checks.
        self.validators.append(MaxChoicesValidator(self.max_choices))

    def deconstruct(self):
        """Return enough information to recreate the field as a 4-tuple."""
        name, path, args, kwargs = super().deconstruct()
        kwargs["max_choices"] = self.max_choices
        return name, path, args, kwargs

    def _check_choices(self):
        """Make the `choices` parameter required."""
        if not self.choices:
            return [
                checks.Error(
                    "MultiSelectFields must define a 'choices' attribute.",
                    obj=self,
                    id="fields.E1001",
                )
            ]
        return super()._check_choices()

    def _check_max_choices_attribute(self):
        """Check that max_choices is well configured."""
        if self.max_choices is None:
            return [
                checks.Error(
                    "MultiSelectFields must define a 'max_choices' attribute.",
                    obj=self,
                    id="fields.E1002",
                )
            ]
        if (
            not isinstance(self.max_choices, int)
            or isinstance(self.max_choices, bool)
            or self.max_choices <= 0
        ):
            return [
                checks.Error(
                    "'max_choices' must be a positive integer.",
                    obj=self,
                    id="fields.E1003",
                )
            ]
        # Make sure max_choices is coherent with max_length
        # Calculate the maximum length this number of choices can represent when stored
        # as a comma separated string:
        ordered_choices = sorted([c[0] for c in self.choices], key=lambda c: -len(c))
        max_length = len(",".join(ordered_choices[: self.max_choices]))
        if max_length > (self.max_length or 0):
            error_message = _(
                "Storing {:d} choices could require storing a CharField of up to {:d} "
                "characters. Please reduce 'max_choices' or increase 'max_length'."
            ).format(self.max_choices, max_length)
            return [checks.Error(error_message, obj=self, id="fields.E1003")]

        return []

    def check(self, **kwargs):
        """Add checks on max_choices to the field's checks."""
        res = [*super().check(**kwargs), *self._check_max_choices_attribute()]
        return res

    @staticmethod
    def from_db_value(value, *_args):
        """Convert a database value to a list value."""
        if not value:
            return None if value is None else []
        return list((v.strip() for v in value.split(",")))

    def to_python(self, value):
        """Convert a string value to a list value. Used for deserialization and in clean forms."""
        if not isinstance(value, str) and hasattr(value, "__iter__"):
            return list(value)
        if not value:
            return None if value is None else []
        return list((v.strip() for v in value.split(",")))

    def get_prep_value(self, value):
        """
        Transform the list value to a concatenation of comma separated strings.

        Arguments:
        ----------
        value (List[string]): a list of strings representing the Python representation
            of multiple Char values.

        Returns:
        --------
        string:
            [] > "" (to differentiate from the null value)
            ["en"] > "en"
            ["en", "fr"] > "en,fr"

        """
        if value is None:
            return None
        return ",".join(value)

    def value_to_string(self, obj):
        """Serialize the Python value. We can use existing methods as it is a CharField."""
        value = self.value_from_object(obj)
        return self.get_prep_value(value)

    def validate(self, value, model_instance):
        """
        Validate each value in values and raise a ValidationError if something is wrong.
        """
        if not self.editable:
            # Skip validation for non-editable fields.
            return

        if self.choices and value:
            # Build a set of possible choices
            choices = set()
            for option_key, option_value in self.choices:
                if isinstance(option_value, (list, tuple)):
                    # This is an optgroup, so look inside the group for
                    # options.
                    for optgroup_key, _optgroup_value in option_value:
                        choices.add(optgroup_key)
                else:
                    choices.add(option_key)

            # Search each value in this set
            invalid_choices = [v for v in value if v not in choices]
            if invalid_choices:
                raise exceptions.ValidationError(
                    self.error_message,
                    code="invalid_choices",
                    params={
                        "number": len(invalid_choices),
                        "value": to_sentence(invalid_choices),
                    },
                )

        if value is None and not self.null:
            raise exceptions.ValidationError(self.error_messages["null"], code="null")

        if not self.blank and not value:
            raise exceptions.ValidationError(self.error_messages["blank"], code="blank")

    def formfield(self, **kwargs):
        """Declare the above MultiSelectFormField as form field for this model field."""
        # don't call super, as that overrides default widget if it has choices
        defaults = {
            "required": not self.blank,
            "label": capfirst(self.verbose_name),
            "help_text": self.help_text,
            "choices": self.choices,
            "max_choices": self.max_choices,
        }
        if self.has_default():
            defaults["initial"] = self.get_default()
        defaults.update(kwargs)
        return MultiSelectFormField(**defaults)

    def contribute_to_class(self, cls, name, private_only=False):
        """
        Add a method to the model using this field to enable the common `get_FIELD_display`
        pattern to display the human readable version of a field based on choices.

        We choose to display a sentence of the form: "French, English and German".
        """
        super().contribute_to_class(cls, name, private_only=private_only)
        if self.choices:
            choicedict = dict(self.choices)

            def func(self, *_args):
                """
                we add `*args` because this method is used in a "render_model" template tag
                and it passes a "request" argument when calling the method.
                For more information, see:
                http://docs.django-cms.org/en/latest/how_to/frontend_models.html
                """
                return capfirst(
                    to_sentence(
                        [choicedict.get(value, value) for value in getattr(self, name)]
                    )
                )

            setattr(cls, "get_%s_display" % self.name, func)



"""Make models.py a module containing one file per model as it was getting too long."""
# flake8: noqa
# pylint: disable=wildcard-import

from .blog import *
from .category import *
from .course import *
from .organization import *
from .person import *
from .program import *
from .role import *



# Generated by Django 2.1.5 on 2019-02-09 10:12

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [("cms", "0022_auto_20180620_1551")]

    operations = [
        migrations.CreateModel(
            name="PlainText",
            fields=[
                (
                    "cmsplugin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        related_name="plain_text_plaintext",
                        serialize=False,
                        to="cms.CMSPlugin",
                    ),
                ),
                ("body", models.TextField(verbose_name="plain text")),
            ],
            options={"abstract": False},
            bases=("cms.cmsplugin",),
        )
    ]



# Generated by Django 2.2.7 on 2019-11-19 15:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("section", "0002_add_template_field"),
    ]

    operations = [
        migrations.AlterField(
            model_name="section",
            name="title",
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
    ]



"""Module used by python-social-auth pipeline."""
from social_core.exceptions import AuthAlreadyAssociated, AuthFailed

USER_FIELDS = ["username", "email"]


# pylint: disable=unused-argument,keyword-arg-before-vararg
def get_username(strategy, details, backend, user=None, *args, **kwargs):
    """Check if the username already exists. Raise an exception if yes."""
    if "username" not in backend.setting("USER_FIELDS", USER_FIELDS):
        return None

    storage = strategy.storage

    if user:
        # The user already exists return its username.
        return {"username": storage.user.get_username(user)}

    email_as_username = strategy.setting("USERNAME_IS_FULL_EMAIL", False)

    if email_as_username and details.get("email"):
        username = details["email"]
    elif details.get("username"):
        username = details["username"]
    else:
        raise AuthFailed(backend, "Failed to retrieve a valid username.")

    if storage.user.user_exists(username=username):
        raise AuthAlreadyAssociated(
            backend, f"user with username {username} already exists."
        )

    return {"username": username}



# Generated by Django 3.1.7 on 2021-02-25 13:35

from django.db import migrations, models

import richie.apps.core.fields.duration


class Migration(migrations.Migration):

    dependencies = [
        ("courses", "0028_auto_20210126_2200"),
    ]

    operations = [
        migrations.RenameField(
            model_name="course",
            old_name="effort",
            new_name="effort_deprecated",
        ),
        migrations.AddField(
            model_name="course",
            name="effort",
            field=richie.apps.core.fields.duration.CompositeDurationField(
                blank=True,
                default_unit="hour",
                help_text="Total amount of time to complete this course.",
                max_length=80,
                null=True,
                time_units={
                    "day": ("day", "days"),
                    "hour": ("hour", "hours"),
                    "minute": ("minute", "minutes"),
                    "month": ("month", "months"),
                    "week": ("week", "weeks"),
                },
            ),
        ),
        migrations.AlterField(
            model_name="course",
            name="duration",
            field=richie.apps.core.fields.duration.CompositeDurationField(
                blank=True,
                default_unit="hour",
                help_text="The course time range.",
                max_length=80,
                null=True,
                time_units={
                    "day": ("day", "days"),
                    "hour": ("hour", "hours"),
                    "minute": ("minute", "minutes"),
                    "month": ("month", "months"),
                    "week": ("week", "weeks"),
                },
            ),
        ),
        migrations.AddField(
            model_name="course",
            name="is_self_paced",
            field=models.BooleanField(
                default=False,
                help_text="Tick if the course pace is self paced.",
                verbose_name="is self paced",
            ),
        ),
    ]
